% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(ace_to_drs, [
		ace_to_drs/3,
		ace_to_drs/5,
		ace_to_drs/6,
		ace_to_drs_time/7
	]).


/** <module> ACE to DRS

@author Kaarel Kaljurand
@version 2008-04-18

*/

:- use_module('../logger/error_logger').

:- use_module('../utils/drs_reverse', [
		drs_reverse/2
	]).

:- use_module(ape_utils, [
		cpu_time/2,
		handle_unknown_words/3
	]).

:- use_module(tokenizer).

:- use_module(tokens_to_sentences, [
		tokens_to_sentences/2
	]).

:- style_check(-singleton).
:- use_module(refres, [
		resolve_anaphors/2
	]).
:- style_check(+singleton).

:- style_check(-discontiguous).
:- style_check(-singleton).
:- use_module('grammar.plp', [parse/3]).
:- style_check(+discontiguous).
:- style_check(+singleton).

%:- debug(result).


%% ace_to_drs(+Text:atom, -Drs:term, -Messages:list) is det.
%% ace_to_drs(+Text:atom, -Sentences:list, -Syntaxtrees:list, -Drs:term, -Messages:list) is det.
%% ace_to_drs(+Text:atom, -Sentences:list, -Syntaxtrees:list, -UnresolvedDrs:term, -Drs:term, -Messages:list) is det.
%
% Provides a simple interface to the parser.
% These predicates will always succeed even if the parser fails.
%
% Example:
%
%==
% ace_to_drs('Every man waits.', Drs, Messages)
% ace_to_drs('Every man waits.', Sentences, Syntaxtrees, Drs, Messages)
%==
%
% @param Text is an ACE text
% @param Sentences is the list of sentences (each a list of tokens) in the ACE text
% @param Syntaxtrees is the list of syntax trees (each a complex list) of the ACE text
% @param UnresolvedDrs is the DRS of the ACE text with anaphoric references unresolved (pre-refres DRS)
% @param Drs is the DRS of the ACE text
% @param Messages is the list of error and warning messages that result from the processing
%
% @see ace_to_drs_time/7 for a slightly different interface to the parser.
%
ace_to_drs(Text, Drs, Messages) :-
	ace_to_drs(Text, _, _, _, Drs, Messages).

ace_to_drs(Text, Sentences, Syntaxtrees, Drs, Messages) :-
	ace_to_drs(Text, Sentences, Syntaxtrees, _, Drs, Messages).

ace_to_drs(Text, Sentences, Syntaxtrees, UnresolvedDrs, Drs, Messages) :-
	clear_ape_messages,
	tokenizer:tokenize(Text, Tokens),

	% BUG: this should be done only if text fails
	ape_utils:get_unknown_words(Tokens, UnknownTokens),
	ape_utils:spellcheck_and_report_unknown_words(UnknownTokens),

	(
		tokens_to_sentences(Tokens, Sentences)
	->
		ignore(grammar:parse(Sentences, Syntaxtrees, UnresolvedUnreversedDrs)),
		ignore(drs_reverse:drs_reverse(UnresolvedUnreversedDrs, UnresolvedDrsPreCopy)),
		copy_term(UnresolvedDrsPreCopy, UnresolvedDrs),
		ignore(refres:resolve_anaphors(UnresolvedDrsPreCopy, DrsTmp)),
		(var(Syntaxtrees) -> Syntaxtrees = []; true),
		(var(DrsTmp) -> Drs = drs([], []); true)
	;
		last(Tokens, LastToken),
		Sentences = Tokens,
		add_error_message(sentence, '', LastToken, 'Every ACE text must end with . or ? or !.')
	),
	(
		is_error_message(_, _, _, _)
	->
		Drs = drs([], [])
	;
		Drs = DrsTmp
	),
	get_messages(Messages),
	!.


%% ace_to_drs_time(+Text, +Parameters:list, -Sentences:list, -Syntaxtrees:list, -Drs:term, -Messages:list, -Times:list) is det.
%
% ace_to_drs_time/7 provides an interface to the parser.
% It includes a try-catch and support for timing. Also, one can specify
% if the word class of unknown words should be guessed or not.
% Note that this interface does not return the pre-refres DRS.
% This predicate will always succeed even if the parser fails.
%
% Example:
% 
%==
% ace_to_drs_time('Every man waits.', [guess:off], Sentences, Syntaxtrees, Drs, Messages, [DT, DP, DR])
%==
%
% @param Text is an ACE text
% @param Parameters is a list of boolean (on/off) parameters
% @param Sentences is the list of sentences (each a list of tokens) in the ACE text
% @param Syntaxtrees is the list of syntax trees (each a complex list) of the ACE text
% @param Drs is the DRS of the ACE text
% @param Messages is the list of error and warning messages that result from the processing
% @param Times is the list of timing info for: tokenizer, parser, refres
%
% @bug clean up the code
% @bug use the standard SWI options format, e.g. [guess(false)]
%
% @see ace_to_drs/[3,5,6] for a slightly different interface to the parser.
%
ace_to_drs_time(
		Text, [guess:GuessOnOff],
		Sentences, Syntaxtrees, Drs, Messages, [DTokenizer, DParse, DRefres]
	) :-
	clear_ape_messages,
	catch(
		(
			cpu_time(ace_to_drs:wrapper_tokenizer(Text, GuessOnOff, Tokens, Sentences), DTokenizer),
			cpu_time(ace_to_drs:wrapper_parser(Sentences, Syntaxtrees, UnresolvedDrs), DParse),
			cpu_time(refres:resolve_anaphors(UnresolvedDrs, DrsTmp), DRefres),
			(var(Syntaxtrees) -> Syntaxtrees = [] ; true),
			(var(DrsTmp) -> Drs = drs([], []) ; true),
			(is_error_message(_, _, _, _) -> Drs = drs([], []) ; Drs = DrsTmp),
			!
		),
		CatchType,
		(
			Sentences = [],
			Syntaxtrees = [],
			Drs = drs([], []),
			DTokenizer = -1,
			DParse = -1,
			DRefres = -1,
			add_error_message(ape, '', CatchType, 'Fatal error. Please send screenshot to APE developers.')
		)
	),
	debug(result, "~w~n", [[Text, GuessOnOff, Tokens, Sentences, Syntaxtrees, DrsTmp, [DTokenizer, DParse, DRefres]]]),
	get_messages(Messages).


% Note: should not fail.
% If sentence splitting fails then the problem must have been that there
% was no sentence end symbol. In this case we return the original
% token list.
wrapper_tokenizer(Text, GuessOnOff, NewTokens, Sentences) :-
	tokenizer:tokenize(Text, Tokens),
	ape_utils:handle_unknown_words(GuessOnOff, Tokens, NewTokens),
	(
		tokens_to_sentences:tokens_to_sentences(NewTokens, SentencesTmp)
	->
		Sentences = SentencesTmp
	;
		Sentences = NewTokens,
		last(NewTokens, LastToken),
		add_error_message(sentence, '', LastToken, 'Every ACE text must end with . or ? or !.')
	).

wrapper_parser(Sentences, Syntaxtrees, DrsReversed) :-
	ignore(grammar:parse(Sentences, Syntaxtrees, Drs)),
	drs_reverse:drs_reverse(Drs, DrsReversed).


clear_ape_messages :-
	clear_messages(character),
	clear_messages(word),
	clear_messages(sentence),
	clear_messages(syntax),
	clear_messages(ape),
	clear_messages(anaphor).
