% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(sentence_failure, [
		get_all_sentence_errors/2, % +SentenceTokens, -ErrorTextList
		get_sentence_error_text/2  % +SentenceTokens, -ErrorText
	]).

:- use_module('../lexicon/lexicon_interface').

:- use_module('../lexicon/chars', [
		is_capitalized/1
	]).

:- use_module('../lexicon/is_in_lexicon', [
		is_functionword/1
	]).

:- use_module('../lexicon/illegalwords').

:- style_check(-singleton).
:- style_check(-discontiguous).
:- use_module('grammar.plp').
:- style_check(+discontiguous).
:- style_check(+singleton).

/** <module> Sentence Failure

@author Norbert E. Fuchs
@author Tobias Kuhn
@author Kaarel Kaljurand
@version 2008-01-25

@bug turn all tokens into lowercase in the beginning, maybe
@bug rules should not fire on illegal words in case those word are quoted (maybe strip quoted strings in the beginning)
@bug In some cases, it is easy to return the token ID, by using nth1/3 instead of member/2

@bug: Add message for 'most' + plural noun (most men are rich)
*/

%% get_all_sentence_errors(+SentenceTokens:list, -ErrorTextList:list) is det.
%
% @param SentenceTokens is a list of tokens in the sentence
% @param ErrorTextList is a list of error messages
%
get_all_sentence_errors(SentenceTokens, ErrorTextList) :-  
	findall(ErrorText, get_sentence_error_text(SentenceTokens, ErrorText), ErrorTextListIntermediate),
	% remove duplicate messages
	list_to_set(ErrorTextListIntermediate, ErrorTextList). 


%% get_sentence_error_text(+SentenceTokens:list, -ErrorText:atom) is nondet.
%
% @param SentenceTokens is a list of tokens in the sentence
% @param ErrorText is an error message
%

% check for illegal words
get_sentence_error_text(SentenceTokens, ErrorText) :-
	member(IllegalWord, SentenceTokens),
	% BUG: add: \+ is_contentword(IllegalWord)
	% because maybe the user defined `any' as a contentword
	is_illegalword(IllegalWord, ErrorText).

% check for repetitions of tokens
get_sentence_error_text(SentenceTokens, ErrorText) :-
    append(_, [Token, '<>', Token|_], SentenceTokens),
	\+ is_repeatable(Token),
	with_output_to(atom(ErrorText), format("Token \'~w\' repeated.", [Token])).

% BUG: experimental: check for 2 function words next to each other
/*
get_sentence_error_text(SentenceTokens, ErrorText) :-
	nth1(Index, SentenceTokens, Token1),
	is_functionword(Token1),
	NextIndex is Index + 1,
	nth1(NextIndex, SentenceTokens, Token2),
	is_functionword(Token2),
	with_output_to(atom(ErrorText), format("Functionwords (\'~w\' and \'~w\') next to each other at position ~w.", [Token1, Token2, Index])).
*/

% check for construct "to + verb"
get_sentence_error_text(SentenceTokens, ErrorText) :-
	(
		append(_, [Have, '<>', to, Verb | _], SentenceTokens)
	;
		append(_, [Have, '<>', to, v, ':', Verb | _], SentenceTokens)
	;
		append(_, [Have, to, '<>', Verb | _], SentenceTokens)
	;
		append(_, [Have, to, '<>', v, ':', Verb | _], SentenceTokens)
	),
	\+ Have = have,
	\+ Have = has,
	(
		iv_finsg(_, Verb)
	;
		tv_finsg(_, Verb)
	;
		tv_finsg(_, Verb)
	;
		tv_pp(_, Verb)
	;
		dv_finsg(_, Verb, _)
	;
		dv_pp(_, Verb, _)
	),
	ErrorText = 'The construct \'to\' + a verb is not allowed. Please replace it by \'that\' + a subordinated sentence.'.

% check for `there is' + proper name
get_sentence_error_text(SentenceTokens, ErrorText) :-
	append(_, [There, is, Token, '<>' | _], SentenceTokens),
	(
		There = there
	;
		There = 'There'
	),
	(
		Token = the
	;
		is_capitalized(Token)
	;
		pn_sg(Token, _, _)
	),
	with_output_to(atom(ErrorText), format("The construct \'there is\' + \'~w\' is not allowed.", [Token])).

% check for `there are' + proper name
get_sentence_error_text(SentenceTokens, ErrorText) :-
	append(_, [There, are, Token, '<>' | _], SentenceTokens),
	(
		There = there
	;
		There = 'There'
	),
	(
		Token = the
	;
		is_capitalized(Token)
	;
		pn_pl(Token, _, _)
	),
	with_output_to(atom(ErrorText), format("The construct \'there are\' + \'~w\' is not allowed.", [Token])).

% check for intransitive verb followed by that-subordination
% Example: John appears that Mary waits. (`appear' is an intransitive verb that is not transitive)
get_sentence_error_text(SentenceTokens, ErrorText) :-
	append(_, [Wordform, '<>', that | _], SentenceTokens),
	(
		iv_finsg(Wordform, Verb)
	;
		iv_infpl(Wordform, Verb)
	),
	with_output_to(atom(ErrorText), format("The intransitive verb \'~w\' cannot be followed by that-subordination. Use a transitive verb.", [Verb])).

% check for missing determiners, beginning of the sentence
get_sentence_error_text(SentenceTokens, ErrorText) :-
	append(['<>', FirstWordOfNBar | NBar], _, SentenceTokens),
	grammar:nbar(_, [FirstWordOfNBar | NBar], _),
	with_output_to(atom(ErrorText), format("Add a determiner before \'~w\'", [FirstWordOfNBar])).

% check for missing determiners, middle of sentence
get_sentence_error_text(SentenceTokens, ErrorText) :-
	append(_, [Wordform, '<>', FirstWordOfNBar | NBar], SentenceTokens),
	FirstWordOfNBar \= does,
	grammar:nbar(_, [FirstWordOfNBar | NBar], _),
	Wordform \= '"',
	Wordform \= but,
	\+ determiner(Wordform),
	\+ lexicon_interface:attributive_adjective(Wordform),
	\+ verb(FirstWordOfNBar),
	with_output_to(atom(ErrorText), format("Add a determiner before \'~w\'", [FirstWordOfNBar])).

get_sentence_error_text(SentenceTokens, 'The sentence contains \'then\' but not \'if\'.') :-
	member(then, SentenceTokens),
	\+ member(if, SentenceTokens),
	\+ member('If', SentenceTokens).

get_sentence_error_text(SentenceTokens, 'The sentence contains \'if\' but not \'then\'.') :-
	(
		member(if, SentenceTokens)
	;
		member('If', SentenceTokens)
	),	
	\+ member(then, SentenceTokens).


% check for illegal use of commas
get_sentence_error_text(SentenceTokens, 'Commas must be followed by \'and\' or \'or\', or must occur at specified positions in lists, sets and commands.') :-
append(Front, [',', NextToken | Tail], SentenceTokens),
	% comma not followed by 'or'
	NextToken \= and,
	% comma not followed by 'and'
	NextToken \= or,
	% comma not within list; not absolutely fool-proof check since illegal comma in "[...] ... , ... [...]" will not be found
	\+ (member('[', Front), member(']', Tail)),
	% comma not within set; not absolutely fool-proof check since illegal comma in "{...} ... , ... {...}" will not be found
	\+ (member('{', Front), member('}', Tail)).
	

get_sentence_error_text(_, 'This is the first sentence that was not ACE. The sign <> indicates the position where parsing failed.').


%%% NEF: Possible extensions
%%% As I see it, checking for missing determiners could be done by my proposal.

%%% if append(L1, [X, Y | L2], ListOfTokensOfSentence)  & noun(Y) & not determiner(X)  then error message "missing determiner"

%%% There is a dog that eats and barks.

%%% This problems occurs only after "there is/are". Thus one might try to find a pattern matching rule for this that recursively calls APE.

%%% if ListOfTokenOfSentence= ([there, is, NP, that, VP1, and, VP2] and np(NP) & vp(VP1)  & vp(VP2) then error message "missing that".


% BUG: some helper predicates, move somewhere else
%
% BUG: we used to have: grammar_functionwords:determiner(_, [WordForm], []).
% but this didn't capture numbers.
% Now we use specifier/3 but this is not complete either as it does not
% capture multiword specifiers as we give a single-element list as an argument.
% In a way we have a solution (possibly incomplete and incorrect) that
% checks only the last token of the determiner, i.e. "at most 2" is a multi-word
% determiner but we only check the "2" which fortunately is also a determiner.
% As would not work for the Saxon genitive determiners (John's, the apples'), we
% add 2 artificial determiners 's' and '\''.
%
determiner('\'') :- !.
determiner(s) :- !.
determiner(WordForm) :-
	grammar:specifier(_, [WordForm], []).

% BUG: this was meant to capture multiword specifiers but
% didn't work, looped instead.
% determiner(WordForm) :-
%	grammar:specifier(_, WordForms, []),
%	last(WordForms, WordForm).


%% is_repeatable(?Token) is nondet.
%
% @param Token is an ACE token that can be repeated
%
% Example: ((1+2)-3) = 0.
%
is_repeatable('{').
is_repeatable('}').
is_repeatable('(').
is_repeatable(')').
is_repeatable('[').
is_repeatable(']').
is_repeatable('"').
