% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(tokenizer, [tokenize/2]).

:- use_module('../lexicon/chars', [
		is_digit/1,
		is_lowercase/1,
		is_uppercase/1,
		is_unused/1
	]).

:- use_module('../logger/error_logger', [
		add_error_message_once/4,
		add_warning_message_once/4
	]).

:- use_module('../utils/ace_niceace', [
		pronoun_split/2
	]).


/** <module> APE Tokenizer

@author Stefan Hoefler 2003-01-28, 2004-02-17
@author Tobias Kuhn 2004-05-17, 2007-10-15
@author Kaarel Kaljurand 2005-10-26, 2006-08-01, 2007-01-08, 2007-01-23
@version 2008-03-28

Comments:

* Strings (between double quotes) are tokenized as one token.
* Saxon Genitives are tokenized as [noun, ', s] and [nouns, ']

* Digits cannot start a token that contains other symbols than digits. Consider
the following input string and its tokenization.

==
123man man123 123man123 man123man

[123, man, man123, 123, man123, man123man]
==
*/


%% tokenize(+ACEText:term, -Tokens:list) is det.
%
% Breaks the ACEText (either atom, or a list of character codes) into a list of tokens (atoms).
% ACEText can either be an atom like 'this is an example' or a list of character
% codes like [116,104,105,115,32,105,...] (or written as "this is an example").
%
% Tokenization will never fail. In case something goes wrong (e.g. a string is
% not closed) then an empty tokenlist is returned.
%
% @param ACEText is the input text, it is either an atom or a string
% @param Tokens is a list of tokens, i.e. the tokenization of the input text
%
tokenize([], []) :- !.

tokenize(Text, Tokens) :-
	atom(Text),
	!,
	atom_codes(Text, CodeList),
	tokenize(CodeList, Tokens).

tokenize([H | Tail], Tokens) :-
	tokenize_x([H | Tail], Tokens).


tokenize_x(CodeList, Tokens) :-
	remove_comments(CodeList, CodeListWithNoComments),
	codes_to_tokens(CodeListWithNoComments, Tokens),
	!.

tokenize_x(_, []).


%% codes_to_tokens(+Codes:list, -Tokens:list) is det.
%
% Merges certain code sequences into single tokens (atoms).
% Later modifies certain token sequences, e.g.
% [can, \', t] -> [can, not].
%
% @param Codes is a list of character codes
% @param Tokens is a list of tokens, i.e. the tokenization of the input text
%
codes_to_tokens(Codes, Tokens) :-
	get_atomics(Codes, TokensTmp),
	expand_contracted_forms(TokensTmp, Tokens).


%% get_atomics(+Codes:list, -Tokens:list) is det.
%
% @param Codes is a list of character codes
% @param Tokens is a list of tokens, i.e. the tokenization of the input text
%
get_atomics([], []).

get_atomics([First | Input], Atomics) :-
	char_type(main, First, FirstT, FirstC),
	complete_line(Input, FirstC, FirstT, Atomics).


%% complete_line(+Input, +FirstC, +FirstT, -Atomics) is det.
%
% Given FirstC (the first character) and FirstT (its type), reads
% and tokenizes the rest of the line into atoms and numbers.

% stop at end
complete_line(_, _, end, []) :- !.

% skip blanks
complete_line(Input, _, blank, Atomics) :-
	!,
	get_atomics(Input, Atomics).

% special character
complete_line(Input, FirstC, special, [A|Atomics]) :-
	!,
	name(A, [FirstC]),
	get_atomics(Input, Atomics).

% begin of a word
complete_line(Input, FirstC, alpha, TokenList) :-
	!,
	complete_word(Input, FirstC, alpha, Word, NextC, NextT, NewInput),
	name_num(Word, A),
	(
		pronoun_split(A, (A1, A2))
	->
		TokenList = [A1, A2 | Atomics]
	;
		TokenList = [A | Atomics]
	),
	complete_line(NewInput, NextC, NextT, Atomics).

% begin of a positive number
complete_line(Input, FirstC, number, [A|Atomics]) :-
	!,
	complete_number(Input, FirstC, number, Word, NextC, NextT, NewInput),
	name_num(Word, A),
	complete_line(NewInput, NextC, NextT, Atomics).

% double quotes that are immediately followed by double quotes
complete_line([34, First | Input], 34, doublequotes, ['"', '"' | Atomics]) :-
	!,
	char_type(_, First, FirstT, FirstC),
	complete_line(Input, FirstC, FirstT, Atomics).

% double quotes
complete_line([First|Input], 34, doublequotes, ['"', A, '"'|Atomics]) :-
	!,
	char_type(string, First, FirstT, FirstC),
	complete_string(Input, FirstC, FirstT, String, NextC, NextT, NewInput),
	name_num(String, A),
	complete_line(NewInput, NextC, NextT, Atomics).


%% complete_word(+Input, +FirstC, +FirstT, -List, -FollC, -FollT, -NewInput)
%
% Given FirstC (the first character) and FirstT (its type),
% reads the rest of a word, putting its characters into List.

complete_word([Next|Input], FirstC, alpha, [FirstC|List], FollC, FollT, NewInput) :-
	!,
	char_type(main, Next, NextT, NextC),
	complete_word(Input, NextC, NextT, List, FollC, FollT, NewInput).

complete_word([], FirstC, alpha, [FirstC], _, end, []) :-
	!.

complete_word([Next|Input], FirstC, number, [FirstC|List], FollC, FollT, NewInput) :-
	!,
	char_type(main, Next, NextT, NextC),
	complete_word(Input, NextC, NextT, List, FollC, FollT, NewInput).

complete_word([], FirstC, number, [FirstC], _, end, []) :-
	!.

% hyphens (= minus signs) are allowed inside of word tokens
complete_word([Next|Input], 45, special, [45|List], FollC, FollT, NewInput) :-
	!,
	char_type(main, Next, NextT, NextC),
	complete_word(Input, NextC, NextT, List, FollC, FollT, NewInput).

complete_word([], 45, special, [45], _, end, []) :-
	!.

complete_word(Input, FirstC, FirstT, [], FirstC, FirstT, Input).


%% complete_number(+Input, +FirstC, +FirstT, -List, -FollC, -FollT, -NewInput)
%
% The same as complete_word, but for numbers.

complete_number([Next|Input], FirstC, number, [FirstC|List], FollC, FollT, NewInput) :-
	!,
	char_type(main, Next, NextT, NextC),
	complete_number(Input, NextC, NextT, List, FollC, FollT, NewInput).

complete_number([], FirstC, number, [FirstC], _, end, []) :-
	!.

% Dots are allowed inside of number tokens if preceded and followed by a digit.
% Note: There is no check that a number token has at most one dot.
complete_number([Next|Input], 46, special, [46|List], FollC, FollT, NewInput) :-
	char_type(main, Next, number, NextC),
	!,
	complete_number(Input, NextC, number, List, FollC, FollT, NewInput).

complete_number(Input, FirstC, FirstT, [], FirstC, FirstT, Input).


%% complete_string(+Input, +FirstC, +FirstT, -String, -NextC, -NextT, -NewInput)
%
% The same as complete_word, but for strings.

complete_string([], _, _, [], _, _, _) :-
	add_error_message_once(character, '', '"', 'Every ACE sentence must contain an even number of quotation marks.').

complete_string([Next|Input], _, doublequotes, [], NextC, NextT, Input) :-
	!,
	char_type(string, Next, NextT, NextC).

complete_string([Next|Input], FirstC, _, [FirstC|List], FollC, FollT, NewInput) :-
	!,
	char_type(string, Next, NextT, NextC),
	complete_string(Input, NextC, NextT, List, FollC, FollT, NewInput).


%% char_type(+Context, +Code, ?Type, -NewCode) is det.
%
% Given an ASCII code, classifies the character as:
% 'blank', 'alpha', 'number', 'special', 'doublequotes'
% and changes it to a potentially different character (NewCode).
%
% BUG: rename: doublequotes -> quote
% BUG: get rid of the Context-arguments

% get0 end of file code
char_type(_, -1, end, -1) :- !.

% replace end of line mark by blank
char_type(_, 10, blank, 32) :- !.

% blank
char_type(_, 32, blank, 32) :- !.

% ctrl codes, etc are treated as blanks
char_type(_, Code, blank, 32) :-
	is_unused(Code), !.

% double quotes
char_type(_, 34, doublequotes, 34) :- !.

% dollar
char_type(_, 36, alpha, 36) :- !.

% degree sign (for "�C")
% BUG: 00A1 (161) is INVERTED EXCLAMATION MARK
% BUG: 00B0 (176) is DEGREE SIGN
char_type(_, 161, alpha, 161) :- !.
char_type(_, 176, alpha, 176) :- !.

% underscore
char_type(_, 95, alpha, 95) :- !.

% letters
char_type(_, Code, alpha, Code) :-
	is_lowercase(Code), !.
char_type(_, Code, alpha, Code) :-
	is_uppercase(Code), !.

% digits
char_type(_, Code, number, Code) :-
	is_digit(Code), !.

% In strings, special characters are kept.
char_type(string, Code, special, Code) :- !.

% The following characters are used in ACE. There others will be filtered out and
% warned about. Note that we apply some additional contextual restrictions, but
% on a higher level than the tokenizer. E.g. slash can only be inside certain
% pronouns and comma can only precede 'and' and 'or'.

% period, question mark, exclamation mark
char_type(_, 46, special, 46) :- !.
char_type(_, 63, special, 63) :- !.
char_type(_, 33, special, 33) :- !.

% hyphen
char_type(_, 45, special, 45) :- !.

% colon (for prefix words' support)
char_type(_, 58, special, 58) :- !.

% apostroph (for Saxon genitive support)
char_type(_, 39, special, 39) :- !.

% slash (for him/her support)
char_type(_, 47, special, 47) :- !.

% comma (for comma-and, comma-or support)
char_type(_, 44, special, 44) :- !.

% plus sign
char_type(_, 43, special, 43) :- !.

% star
char_type(_, 42, special, 42) :- !.

% parentheses ()
char_type(_, 40, special, 40) :- !.
char_type(_, 41, special, 41) :- !.

% square brackets []
char_type(_, 91, special, 91) :- !.
char_type(_, 93, special, 93) :- !.

% curly bracktes {}
char_type(_, 123, special, 123) :- !.
char_type(_, 125, special, 125) :- !.

% < = > \
char_type(_, 60, special, 60) :- !.
char_type(_, 61, special, 61) :- !.
char_type(_, 62, special, 62) :- !.
char_type(_, 92, special, 92) :- !.

% ampersand
char_type(_, 38, special, 38) :- !.


% Remove all strange characters. This should make the tokenizer more robust.
% BUG: test comment-support
% BUG: This filters out all weird characters, e.g. Japanese
% proper names would not be supported.

char_type(main, Code, blank, 32) :-
	%char_code(Char, Code),
	with_output_to(atom(CharCode), format("~c (0x~16r, ~10r)", [Code, Code, Code])),
	add_warning_message_once(character, '', CharCode, 'Unknown character(s) ignored.').


%% name_num(+Codes:list, -AtomOrNumber:atomic) is det.
%
% Converts a list of character codes into an atom or a number.
%
% @param Codes is a list of character codes
% @param AtomOrNumber is an atom or a number
%
% @bug rename this predicate
%
name_num(Codes, AtomOrNumber) :-
	atom_codes(Atom, Codes),
	catch( atom_number(Atom, AtomOrNumber), _, AtomOrNumber = Atom ).


%% remove_comments(+Tokens:list, -NewTokens:list) is det.
%% remove_comments(+Tokens:list, +OnOff:atom, -NewTokens:list) is det.
%
% Remove C and Perl comments, where
%
% * 35 is #
% * 10 is \n
% * 47 is /
% * 42 is *
%
% * off - comment-environment off.
% * con - C-comment-environment on.
% * on  - Perl-comment-environment on.

remove_comments(OldList, NewList) :-
	remove_comments(OldList, off, NewList).


% Remove starting C-comments. Keep Perl-comment-environment on.
remove_comments([47, 42 | OldTail], on, NewTail) :- !, remove_comments(OldTail, on, NewTail).

% Remove starting C-comments. Switch C-comment-environment on.
remove_comments([47, 42 | OldTail], _, NewTail) :- !, remove_comments(OldTail, con, NewTail).

% Remove ending C-comments. Keep Perl-comment-environment on.
remove_comments([42, 47 | OldTail], on, NewTail) :- !, remove_comments(OldTail, on, NewTail).

% Remove ending C-comments. Switch comment-environment off.
remove_comments([42, 47 | OldTail], _, NewTail) :- !, remove_comments(OldTail, off, NewTail).

% Remove Perl-comment. C-comment-environment stays on.
remove_comments([35 | OldTail], con, NewTail) :- !, remove_comments(OldTail, con, NewTail).

% Remove Perl-comment. Switch Perl-comment-environment on.
remove_comments([35 | OldTail], _, NewTail) :- !, remove_comments(OldTail, on, NewTail).

% Remove newline. C-comment-environment stays on.
remove_comments([10 | OldTail], con, NewTail) :- !, remove_comments(OldTail, con, NewTail).

% Remove newline. Switch comment-environment off.
remove_comments([10 | OldTail], _, [10 | NewTail]) :- !, remove_comments(OldTail, off, NewTail).

% Remove character. C-comment-environment stays on.
remove_comments([_ | OldTail], con, NewTail) :- remove_comments(OldTail, con, NewTail).

% Remove character. Perl-comment-environment stays on.
remove_comments([_ | OldTail], on, NewTail) :- remove_comments(OldTail, on, NewTail).

% Preserve character.
remove_comments([H | OldTail], off, [H | NewTail]) :- remove_comments(OldTail, off, NewTail).

remove_comments([], _, []).


%% expand_contracted_forms(+TokenListIn:list, -TokenListOut:list) is det.
%
% @bug: `cannot' could also be handled during (pronoun) splitting
%
expand_contracted_forms([], []).

expand_contracted_forms(['No', one | RestIn], ['No', '-one' | RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([no, one | RestIn], [no, '-one' | RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([isn, '\'', t|RestIn], [is, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([aren, '\'', t|RestIn], [are, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([doesn, '\'', t|RestIn], [does, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([don, '\'', t|RestIn], [do, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([can, '\'', t|RestIn], [can, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([cannot|RestIn], [can, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([shouldn, '\'', t|RestIn], [should, not|RestOut]) :-
	!,
	expand_contracted_forms(RestIn, RestOut).

expand_contracted_forms([Token|TailIn], [Token|TailOut]) :-
	expand_contracted_forms(TailIn, TailOut).
