% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(tokens_to_sentences, [
		tokens_to_sentences/2
	]).


/** <module> APE Sentence splitter

Converts a flat list of tokens into a list of sentences,
each of which is a list of tokens. Sentences end with
one of the three symbols: '.', '?', and '!'.

For example, the following list of tokens:

==
[John, likes, Mary, ., Every, man, owns, a, car, .]
==

is converted into the following list of sentences

==
[[John, likes, Mary, .], [Every, man, owns, a, car, .]]
==

@author Kaarel Kaljurand
@version 2008-03-28

@bug should we generate error messages here (if token list does not end with ./?/!)
@bug maybe move is_sentence_end_symbol/1 into '../lexicon/chars.pl'

*/


%% tokens_to_sentences(+Tokens:list, -Sentences:list) is semidet.
%
% Succeeds if Tokens is a list of ACE sentences, in this case
% the sentences are returned. To succeed the token list must either
% be empty or end with a sentence end symbol. E.g. the following
% token list would cause a failure:
%
%==
%[a, b, c]
%==
%
% @param Tokens is a list of ACE tokens
% @param Sentences is a list of sentences (where a sentence is a list of ACE tokens)
%
tokens_to_sentences([], []).

tokens_to_sentences(Tokens, [SentenceOut|Sentences]) :-
	first_sentence(Tokens, Sentence, RestTokens),
	add_sentence_beginning_token(Sentence, SentenceOut),
	tokens_to_sentences(RestTokens, Sentences).


%% add_sentence_beginning_token(+SentenceIn, -SentenceOut)
%
add_sentence_beginning_token([Prop,Var,':'|Rest], [Prop,Var,':',^|Rest]) :-
    (Prop == 'Proposition' ; Prop == proposition ; Prop == 'Fact' ; Prop == fact ),
    !.

add_sentence_beginning_token(Sentence, [^|Sentence]).


%% first_sentence(+Tokens:list, -Sentence:list, -RestTokens:list) is det.
%
% Note that the token list can contain variables (added by the guesser
% to match prefixed in the parser). Thus we have to make sure that a variable
% is not "mistaken" for a sentence end symbol.
%
% @param Tokens is a list of ACE tokens
% @param Sentence is an ACE sentence (a list of ACE tokens)
% @param RestTokens is a list of ACE tokens
%
first_sentence([SentenceEndSymbol | RestTokens], [SentenceEndSymbol], RestTokens) :-
	nonvar(SentenceEndSymbol),
	is_sentence_end_symbol(SentenceEndSymbol),
	!.

first_sentence([Token | RestTokens], [Token | RestSentence], RestTokens2) :-
	first_sentence(RestTokens, RestSentence, RestTokens2).


%% is_sentence_end_symbol(?Token:atom) is semidet.
%
% Tests if Token is an ACE sentence end symbol.
%
% @param Token is a token.
%
is_sentence_end_symbol('.').
is_sentence_end_symbol('?').
is_sentence_end_symbol('!').
