% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_to_owldrs, [
		drs_to_owldrs/2
	]).

/** <module> DRS to OWL DRS converter

The DRS is converted to "OWL DRS" which is a more suitable format
for the conversion into OWL.
(The input must be simplified DRS (drs_to_sdrs.pl) which has been numbervard.)

1. We rewrite implication chains A -> (B -> C) as (A & B) -> C.

2. We rewrite embedded implications via conjunction and negations.

3. We remove double negations.

4. We rewrite certain conditions and sets of conditions:
* transitive adjectives are turned into verbs,
* relation/3 + be-predicate are used to construct data properties,
* otherwise, relation/3 is used to construct object properties.

5. We convert the new condition type IfBox => [quantity/3] (introduced with ACE 6)
into a negation box to obtain a better compatibility with the PhD thesis of Kaarel.

@author Kaarel Kaljurand
@version 2008-06-30

TODO
- Get rid of named-objects here, so that we would not have to deal with them
later. For each named-object(A, 'John'), unify A = named('John')
*/


% Operators used in the DRS.
:- op(400, fx, -).
:- op(400, fx, ~).
:- op(400, fx, <>).
:- op(400, fx, []).
:- op(500, xfx, =>).
:- op(500, xfx, v).



%% drs_to_owldrs(+Drs:term, -OwlDrs:term) is det.
%
% Modifies the DRS to make it easier to convert in to OWL/SWRL.
%
% @param Drs is Attempto DRS
% @param OwlDrs is a modified Attempto DRS
%
drs_to_owldrs(Drs, OwlDrs) :-
	drs_to_owldrs(0, Drs, OwlDrs).


%% drs_to_owldrs(+Level:term, +Drs:term, -OwlDrs:term) is det.
%
% Modifies the DRS to make it easier to convert in to OWL/SWRL.
% The embedding-level is 0 for the top-level DRS-box, and grows
% by 0, l(0), l(l(0)), ...
%
% @param Level denotes the level of embedding of the DRS-box
% @param Drs is Attempto DRS
% @param OwlDrs is a modified Attempto DRS
%
drs_to_owldrs(_, [], []) :-
	!.

% Double negation is removed
% --A == A
drs_to_owldrs(L, [-Drs | Conds], CondsAll) :-
	drs_to_owldrs(l(L), Drs, [-DrsR]),
	!,
	drs_to_owldrs(L, Conds, CondsR),
	append(DrsR, CondsR, CondsAll).

drs_to_owldrs(L, [-Drs | Conds], [-DrsR | CondsR]) :-
	!,
	drs_to_owldrs(l(L), Drs, DrsR),
	drs_to_owldrs(L, Conds, CondsR).

drs_to_owldrs(L, [<>(Drs) | Conds], [<>(DrsR) | CondsR]) :-
	!,
	drs_to_owldrs(l(L), Drs, DrsR),
	drs_to_owldrs(L, Conds, CondsR).

drs_to_owldrs(L, [[](Drs) | Conds], [[](DrsR) | CondsR]) :-
	!,
	drs_to_owldrs(l(L), Drs, DrsR),
	drs_to_owldrs(L, Conds, CondsR).

drs_to_owldrs(L, [~(Drs) | Conds], [~(DrsR) | CondsR]) :-
	!,
	drs_to_owldrs(l(L), Drs, DrsR),
	drs_to_owldrs(L, Conds, CondsR).


% Handle the IfBox => [quantity/3] condition by converting it into negation box.
% This preserves the code in drs_to_owlswrl.pl as it was presented in Kaarel's PhD thesis.
% In the future we might handle the new IfBox => [quantity/3] directly in drs_to_owlswrl.pl
% as it gives a more direct mapping (no need for negation). On the other hand
% it seems like more work.
drs_to_owldrs(L, [IfBox => [quantity(Ref, QType, QNum)-Id] | Conds], DrsR) :-
	!,
	select(object(Ref, Lemma, Count, na, na, na)-Id, IfBox, RemainingIfBox),
	negate_quantifier(QType, NegQType),
	drs_to_owldrs(L, [-[object(Ref, Lemma, Count, na, NegQType, QNum)-Id | RemainingIfBox] | Conds], DrsR).

% Handle sentences like:
% For every thing X for every thing Y if X owns something that contains Y then X owns Y.
% BUG: maybe allow this for any level, i.e. not just 0
% BUG: this new handling changes the OWL representation of sentences 41, 115, 116, 117,
% Also, it makes sentences 66 and 67 correctly succeed as SRWL rules.
% Also, it would break sentence 57 "Every man likes no dog.", but we avoid this
% by requiring the embedded then-box not to be negated.
% But, then we reject this: For everything X for every thing Y if X likes Y then X does not hate Y.
% (although this could be mapped to property disjointness axiom).
%
% BUG: Think about all this in a more systematic way. (The rewriting is definitely
% correct, but it might harm completeness.)
drs_to_owldrs(0, [IfBox => [EmbeddedIfBox => EmbeddedThenBox] | Conds], DrsR) :-
	EmbeddedThenBox \= [-_],
	EmbeddedThenBox \= [quantity(_, _, _)-_],
	!,
	append(IfBox, EmbeddedIfBox, MergedIfAndEmbeddedIfBox),
	drs_to_owldrs(0, [MergedIfAndEmbeddedIfBox => EmbeddedThenBox | Conds], DrsR).

%
% Toplevel implications are kept as they are.
%
drs_to_owldrs(0, [Drs1 => Drs2 | Conds], [Drs1R => Drs2R | CondsR]) :-
	!,
	drs_to_owldrs(l(0), Drs1, Drs1R),
	drs_to_owldrs(l(0), Drs2, Drs2R),
	drs_to_owldrs(0, Conds, CondsR).

/*
% This is handled by the double negation removal rule.
% Re-enable it to support: What is a country that borders no sea?
% A => -B == -(A & B)
drs_to_owldrs(l(L), [IfBox => [-ThenBox] | Conds], DrsR) :-
	!,
	append(IfBox, ThenBox, IfThenBox),
	drs_to_owldrs(l(L), [-IfThenBox | Conds], DrsR).
*/


% Embedded implications are turned into double negations.
% A => B == -(A & -B)
drs_to_owldrs(l(L), [IfBox => ThenBox | Conds], DrsR) :-
	!,
	drs_to_owldrs(l(L), [-[-ThenBox | IfBox] | Conds], DrsR).

drs_to_owldrs(L, [DRS1 v DRS2 | Conds], [DRS1R v DRS2R | CondsR]) :-
	!,
	drs_to_owldrs(l(L), DRS1, DRS1R),
	drs_to_owldrs(l(L), DRS2, DRS2R),
	drs_to_owldrs(L, Conds, CondsR).

drs_to_owldrs(L, Conds, CondsR) :-
	rewrite_relation(Conds, FilteredConds),
	drs_to_owldrs(L, FilteredConds, CondsR).

% (Fallback)
drs_to_owldrs(L, [Cond | Conds], [Cond | CondsR]) :-
	drs_to_owldrs(L, Conds, CondsR).


% of-constructions with data items
%
% John's age is 30.
% John's address is "Poland".
%
% BUG: those should be rejected I guess because data properties
% cannot have inverses. (This can be done later, by
% not allowing a data object to be a subject.)
%
% 30 is an age of John.
% "Poland" is an address of John.
%
% Note that in addition to relation/3 we also remove the object/8.
%
rewrite_relation(Conds, [predicate(C, PropertyName, A, DataItem)-Id | FilteredConds]) :-
	remove_many([
		predicate(C, be, D, DataItem)-Id,
		object(D, PropertyName, countable, _, eq, 1)-Id,
		relation(D, of, A)-Id
	], Conds, FilteredConds),
	is_dataitem(DataItem),
	!.

rewrite_relation(Conds, [predicate(C, PropertyName, A, DataItem)-Id | FilteredConds]) :-
	remove_many([
		predicate(C, be, DataItem, D)-Id,
		object(D, PropertyName, countable, _, eq, 1)-Id,
		relation(D, of, A)-Id
	], Conds, FilteredConds),
	is_dataitem(DataItem),
	!.


% of-constructions without data items
%
% We handle differently the sentences with copula and without copula.
% The following sentences
%
% John's father is Bill.
% Bill is a father of John.
%
% will be represented as:
%
% ObjectPropertyAssertion(ObjectProperty(father), Individual(John), Individual(Bill))
% ...
%
% But the following sentences
%
% John's father likes Bill.
% Bill likes John's father.
%
% will be handled by keeping the object/8 in the returned list.
%
% ObjectPropertyAssertion(ObjectProperty(father), Individual(John), AnonymousIndividual($VAR(2)))
% ClassAssertion(AnonymousIndividual($VAR(2)), OWLClass(father))
% ...
%
rewrite_relation(Conds, [predicate(_, PropertyName, A, B)-Id | FilteredConds]) :-
	remove_many([
		predicate(_, be, D, B)-Id,
		object(D, PropertyName, countable, _, eq, 1)-Id,
		relation(D, of, A)-Id
	], Conds, FilteredConds),
	!.

rewrite_relation(Conds, [predicate(_, PropertyName, A, B)-Id | FilteredConds]) :-
	remove_many([
		predicate(_, be, B, D)-Id,
		object(D, PropertyName, countable, _, eq, 1)-Id,
		relation(D, of, A)-Id
	], Conds, FilteredConds),
	!.

rewrite_relation(Conds, [
	predicate(_, PropertyName, A, D)-Id,
	object(D, PropertyName, countable, na, eq, 1)-Id
	| FilteredConds]) :-
	remove_many([
		object(D, PropertyName, countable, _, eq, 1)-Id,
		relation(D, of, A)-Id
	], Conds, FilteredConds).


% predicate/4
%
% Transitive and comparative adjectives (with arguments)
% are mapped to transitive verbs. This is debatable (and not reversible).
% BUG: We do not currently check the Degree of property/4.
% Note that it can be one of {pos, pos_as, comp, comp_than, sup}
% The outcome is that "John is as rich as Mary." == "John is richer than Mary."
% which is not wanted. I.e. one should use different property names and additionally
% assert disjointness of the properties.
%
% Examples:
%
% John is fond-of Mary.        # pos
% England is located-in UK.    # pos
% John is as rich as Mary.     # pos_as
% John is fonder-of Mary.      # comp
% John is taller than Mary.    # comp_than
% John is more rich than Mary. # comp_than
% John is fondest-of Mary.     # sup

rewrite_relation(Conds, [predicate(RefPred, PropertyName, RefArg1, RefArg2)-Id | FilteredConds]) :-
	remove_many([
		predicate(RefPred, be, RefArg1, RefRel)-Id,
		property(RefRel, PropertyName, _Degree, RefArg2)-Id
	], Conds, FilteredConds).


% Which man ... ? --> What is a man that ... ?
% Turns plural into singular.
% BUG: turns mass into countable (can we have 'mass' with 'which' at all?)
rewrite_relation(Conds, [
		query(NewRef1, what)-Id,
		object(WhichRef, Name, countable, na, eq, 1)-Id,
		predicate(_, be, NewRef1, WhichRef)-Id | FilteredConds]) :-
	remove_many([
		query(WhichRef, which)-Id,
		object(WhichRef, Name, _, _, _, _)-Id
	], Conds, FilteredConds).

%% remove_many(+ConditionsToRemove:list, +AllConditions:list, RemainingConditions:list) is det.
%
%
remove_many([], Conds, Conds).

remove_many([H | T], Conds, FilteredConds) :-
	select(H, Conds, NewConds),
	remove_many(T, NewConds, FilteredConds).


%% is_dataitem(+DataItem:term) is det.
%
% Succeeds if DataItem is an ACE int, real, or string.
%
% @param DataItem is an ACE data item (a number or a string)
%
is_dataitem(int(_)).
is_dataitem(real(_)).
is_dataitem(string(_)).


%% negate_quantifier(+QType:atom, -NegQType:atom) is det.
%
% @param QType is a quantifier type from {leq, less}
% @param NegQType is the "opposite" quantifier type (from {greater, geq})
%
negate_quantifier(leq, greater).
negate_quantifier(less, geq).
